/*
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "mime.h"

/* An implementation of some handy RFC 2045, 2046, 2047 MIME-related
 * functions for ezimail.
 */



const char HEX_CHARS[]             = "0123456789ABCDEF";
const char RFC2045_SPECIAL_CHARS[] = "@.,;:<>[]\\\"()?/= \t";
#define    RFC2045_MAX_LINE_LEN 75  /* The max width is really 76 chars, but 
				     * we need to leave space for an extra
				     * character in the code.  This code 
				     * does produce a 76-character line.
				     */
const char RFC2045_NIL[] = "NIL";

/* Check whether the given character is safe to send over mail gateways
 * without being clobbered.  EBCDIC machines are still alive and well.
 */
int rfc2045_is_safe_char ( const char c )
{
  return ( ( ( c >= 33 && c <= 126 ) && 
	     ( c != 61 ) && 
	     ( !strchr ( RFC2045_SPECIAL_CHARS, c ) ) ) ? 1 : 0 );
}

/* Is this whitespace chracter safe to send over a mail gateway? */
int rfc2045_is_safe_ws ( const char c ) 
{
  return ( ( c == ' ' || c == '\t' ) ? 1 : 0 );
}

/* Turn a 2-digit hex value into a chracter */
char rfc2045_hex_to_char ( const char * value )
{
  char
    buf[3];
  
  buf[0] = value[0];
  buf[1] = value[1];
  buf[2] = '\0';

  return (char) strtol ( buf, NULL, 16 );
}

/* Turn a charcter into a 2-digit RFC2045 hex number, need 
 * a buffer of at least three digits 
 */
char * rfc2045_char_to_hex ( const char c, char * buf )
{
  buf[0] = HEX_CHARS[c / 16];
  buf[1] = HEX_CHARS[c % 16];
  buf[2] = '\0';
  return buf + 2;
}


/* This function is an RFC2045-compliant quoted-printable
 * encoder.  It takes a source buffer (null terminated) and 
 * destination buffer, along with the destination buffer's
 * length.
 * 
 * If there is enough buffer space, the length of the destination
 * buffer is returned, otherwise -1 is returned and the function
 * should be called with a larger buffer
 *
 */
int rfc2045_to_mime_quoted_printable ( const char * s,
				       char       * d,
				       int          outLen )
{
  char 
    * inP,
    * outP;

  int
    linePos = 0,
    outPos  = 0;
  
  inP  = (char*)s;
  outP = (char*)d;

  while ( *inP != '\0' ) { 
    
    if ( outPos + 1 >= outLen )
      return -1;

    if ( rfc2045_is_safe_char ( *inP ) ||
	 ( linePos < RFC2045_MAX_LINE_LEN && rfc2045_is_safe_ws ( *inP ) ) ) { 
      *outP++ = *inP++, outPos++;
      linePos++;

      if ( linePos >= RFC2045_MAX_LINE_LEN ) {
	if ( outPos + 3 >= outLen ) 
	  return -1;
	*outP++ = '=', outPos++;
	*outP++ = '\r', outPos++;
	*outP++ = '\n', outPos++;
	linePos = 0;
      }
      
    } else { 

      if ( outPos + 3 >= outLen )
	return -1;

      /* Need to encode it as a =XX character */
      if ( linePos >= ( RFC2045_MAX_LINE_LEN - 2 ) ) {
	*outP++ = '=', outPos++;
	*outP++ = '\r', outPos++;
	*outP++ = '\n', outPos++;
	linePos = 0;
      }
      
      *outP++ = '=', outPos++;
      outP = rfc2045_char_to_hex ( *inP++, outP ), outPos += 2;
      linePos += 3;
    }
  }

  *outP = '\0';
  return outPos;
}

/* Convert a quoted-printable buffer into a normal unix 
 * text buffer that is suitable for editing.
 */
int rfc2045_from_mime_quoted_printable ( const char * in, 
					 char * out,
					 int    outLen )
{
  char 
    * outP,
    * inP;
  
  int
    outPos = 0;
  
  inP  = (char*)in;
  outP = (char*)out;
  
  if ( in == NULL || out == NULL || outLen <= 0 ) 
    return -1;
  
  while ( outPos < outLen && *inP != '\0' ) { 
    while ( *inP != '\0' && *inP != '=' && outPos < outLen && *inP != '\r' ) 
      *outP++ = *inP++, outPos++;

    /* Get rid of trailing whitespace */
    if ( *inP == '\r' ) { 
      char 
	* inP2 = inP - 1;

      while ( isspace ( *inP2-- ) )
	outP--, outLen--;
      inP += 2;
      *outP++ = '\r';
      *outP++ = '\n';
      outLen += 2;

    }

    if ( *inP == '=' ) {
      inP++;
      if ( *(inP) == '\0'|| *(inP+1) == '\0' ) {
	*outP = '\0';
	return outPos;
      }
      
      /* Check for an = at the end of the line */
      if ( !(*inP == '\r' && *(inP+1) == '\n') ) {
	*outP++ = rfc2045_hex_to_char ( inP );
	outPos++;
      }
      inP += 2;
    }
  }

  return outPos;
}

char * rfc2045_mime_headers ( char * type,
			      char * subtype,
			      char * charset,
			      char * encoding,
			      char * buffer,
			      int    bufLen )
{
  int 
    newLen;
  
  /* These are reasonable defaults for very portable ASCII text */
  if ( type == NULL )     type = "text";
  if ( subtype == NULL )  subtype = "plain";
  if ( charset == NULL )  charset = "iso-8859-1";
  if ( encoding == NULL ) encoding = "7bit";

      newLen = snprintf ( buffer,
			  bufLen,
		     "Content-Type: %s/%s;\r\n" 
		     " charset=\"%s\"\r\n"
		     "Content-Transfer-Encoding: %s\r\n\r\n",
		     type, subtype, charset, encoding );

 return ( buffer + newLen );
}


/* Decode the RFC 2060 FETCH x BODYSTRUCTURE response.
 * Mark Crispin must be removed from this planet for having written
 * such an incomprehensible description on this in 7.4.2 FETCH Response.
 *
 * A BNF specification would have made his life much easier and made
 * me much happier.  The guy should be placed in a remedial English 
 * class.  Phew.  I feel better now.
 */

/* Get the next string token */
static char * next_string_token ( char * bs, char * token )
{
  int
    len = 0;

  while ( isspace ( *bs ) )
    bs++;

  /* Get the NIL arguments */
  if ( !strcasecmp ( RFC2045_NIL, bs ) ) { 
    bs += strlen ( RFC2045_NIL );
    strcpy ( token, RFC2045_NIL );
    return bs;
  }

  if ( *bs != '\"' )
    return NULL;
  
  while ( *bs != '\0' && 
	  *bs != '\"' && len < STR255LEN - 1 )
    *token++ = *bs++, len++;
  
  if ( *bs == '\"' )
    return ++bs;
  
  return NULL;
}

/* Get the next integer token */
static char * next_integer_token ( char * bs, int * token )
{
  while ( isspace ( *bs ) )
    bs++;
  
  if ( !isdigit ( *bs ) )
    return NULL;
  
  *token = strtol ( bs, &bs, 10 );
  
  return bs;
}

/* Slurp up an entire parameter list in parens, in case we don't
 * know how to parse it.
 */
static char * next_paren_list_token ( char * bs, char * token )
{
  int 
    parenCounter = 0,
    tokenPos = 0,
    inQuotes = 0;

  /* ( "foo" "bar" ... ), allow nested parens and parens in quotes */
  while ( isspace ( *bs ) )
    bs++;
  
  if ( *bs != '(' )
    return NULL;
  parenCounter++;

  while ( parenCounter > 0 ) { 
    
    while ( *bs != '\0' &&  tokenPos < STR255LEN ) { 

      switch ( *bs ) { 
      case '(':
	if ( !inQuotes ) parenCounter++;
	break;
      case ')':
	if ( !inQuotes ) parenCounter--;
	break;
      case '\"':
	inQuotes = !inQuotes;
	break;
      }
      
      *token++ = *bs++, tokenPos++;

    }
  }

  if ( parenCounter > 0 )
    return NULL;
  else
    return bs;
}

/* Same as above, except also take a single string token */
static char * next_paren_list_or_string_token ( char * bs, char * token )
{
  while ( isspace ( *bs ) )
    bs++;

  if ( *bs == '\"' )
    return next_string_token ( bs, token );
  else if ( *bs == '(' )
    return next_paren_list_token ( bs, token );
  else
    return NULL;
}

/* Retreive the integer constant version of the first field 'type'
 * in the body structure.  Note that this blows away the extension type,
 * but we don't know much about how to support that now anyway...
 */
char * rfc2060_get_type ( char * bs, int * type )
{
  char typeStr[STR255LEN] = { 0 };
  
  bs = next_string_token ( bs, typeStr );
  
  *type = MIME_TYPE_INVALID;
  
  if ( bs == NULL ) 
    return NULL;

  if ( !strcasecmp ( "TEXT",        typeStr ) ) *type = MIME_TYPE_TEXT;
  if ( !strcasecmp ( "IMAGE",       typeStr ) ) *type = MIME_TYPE_IMAGE;
  if ( !strcasecmp ( "AUDIT",       typeStr ) ) *type = MIME_TYPE_AUDIO;
  if ( !strcasecmp ( "VIDEO",       typeStr ) ) *type = MIME_TYPE_VIDEO;
  if ( !strcasecmp ( "APPLICATION", typeStr ) ) *type = MIME_TYPE_APPLICATION;
  if ( !strcasecmp ( "X-",          typeStr ) ) *type = MIME_TYPE_EXTENSION;
  if ( !strcasecmp ( "MESSAGE",     typeStr ) ) *type = MIME_TYPE_MESSAGE;
  if ( !strcasecmp ( "MULTIPART",   typeStr ) ) *type = MIME_TYPE_MULTIPART;

  return bs;
}

char * rfc2060_get_subtype ( char * bs, char * subtype )
{
  return next_string_token ( bs, subtype );
}

char * rfc2060_get_body_parameter ( char * bs, char * params )
{
  return next_paren_list_token ( bs, params );
}

char * rfc2060_get_body_id ( char * bs, char * id )
{
  return next_string_token ( bs, id );
}

char * rfc2060_get_body_description ( char * bs, char * description )
{
  return next_string_token ( bs, description );
}

char * rfc2060_get_body_encoding ( char * bs, char * encoding )
{
  return next_string_token ( bs, encoding );
}

char * rfc2060_get_body_size ( char * bs, int * octets, int * lines )
{
  bs = next_integer_token ( bs, octets );
  if ( bs == NULL ) return NULL;
  
  bs = next_integer_token ( bs, lines );
  if ( bs == NULL ) return NULL;

  return bs;
}

char * rfc2060_get_body_language ( char * bs, char * language )
{
  return next_string_token ( bs, language );
}

char * rfc2060_get_text_body_size ( char * bs, int * lines )
{
  return next_integer_token ( bs, lines );
}

char * rfc2060_get_extension_fields ( char * bs, 
				      char * parameter,
				      char * disposition,
				      char * language )
{
  bs = next_paren_list_or_string_token ( bs, parameter );
  if ( bs == NULL ) return NULL;

  bs = next_string_token ( bs, disposition );
  if ( bs == NULL ) return NULL;
  
  bs = next_string_token ( bs, language );
  if ( bs == NULL ) return NULL;
  
  return bs;
}

MimeAttachmentList * alloc_mal ( void )
{
  MimeAttachmentList 
    * mal;
  mal = malloc ( sizeof(MimeAttachmentList) );
  if ( mal == NULL ) 
    return NULL;
  bzero ( mal, sizeof(MimeAttachmentList) );
  return mal;
}

MimeAttachment * alloc_ma ( void )
{
  MimeAttachment
    * mal;
  mal = malloc ( sizeof(MimeAttachment) );
  if ( mal == NULL ) 
    return NULL;
  bzero ( mal, sizeof(MimeAttachment) );
  return mal;
}

/* The discrete media types are layed out as:
 *  "type" "subtype" (parameter list) "id" "description" 
 *  "encoding" octet_size line_size 
 */

char * rfc2060_decode_discrete ( char * bs, 
				 MimeAttachmentList * mal,
				 int type )
{
  /* Allocate and decode the current discrete type */
  mal->value = alloc_ma();
  if ( mal->value == NULL ) return NULL;

  mal->value->type = type;
  bs = rfc2060_get_subtype ( bs, mal->value->subtype );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_parameter ( bs, mal->value->parameter );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_id ( bs, mal->value->id );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_description ( bs, mal->value->description );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_encoding ( bs, mal->value->encoding );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_size ( bs,
			       &mal->value->sizeOctets,
			       &mal->value->sizeLines );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_extension_fields ( bs, 
				      mal->value->ext_parameter,
				      mal->value->ext_disposition,
				      mal->value->ext_language );
  if ( bs == NULL ) return NULL;

  return bs;  
}

/* Multipart looks something like:
 * "type" "subtype" (parameter list) (disposition list) (language)|"language"
 */
char * rfc2060_decode_multipart ( char * bs,
				  MimeAttachmentList * mal )
{
  mal->value = alloc_ma();
  if ( mal->value == NULL ) return NULL;
  mal->value->type = MIME_TYPE_MULTIPART;

  bs = rfc2060_get_subtype ( bs, mal->value->subtype );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_parameter ( bs, mal->value->ext_parameter );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_parameter ( bs, mal->value->ext_disposition );
  if ( bs == NULL ) return NULL;

  bs = next_paren_list_or_string_token ( bs, mal->value->ext_language );
  if ( bs == NULL ) return NULL;

  return NULL;
}

char * rfc2060_decode_message ( char * bs,
				  MimeAttachmentList * mal )
{
  char 
    MD5[STR255LEN];

    mal->value = alloc_ma();
  if ( mal->value == NULL ) return NULL;

  mal->value->type = MIME_TYPE_MESSAGE;
  bs = rfc2060_get_subtype ( bs, mal->value->subtype );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_parameter ( bs, mal->value->parameter );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_id ( bs, mal->value->id );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_description ( bs, mal->value->description );
  if ( bs == NULL ) return NULL;
  
  bs = rfc2060_get_body_encoding ( bs, mal->value->encoding );
  if ( bs == NULL ) return NULL;

  bs = rfc2060_get_body_size ( bs,
			       &mal->value->sizeOctets,
			       &mal->value->sizeLines );
  if ( bs == NULL ) return NULL;

  bs = next_string_token ( bs, MD5 );
  if ( bs == NULL ) return NULL;

  bs = next_paren_list_or_string_token ( bs, mal->value->ext_disposition );
  if ( bs == NULL ) return NULL;
  
  bs = next_paren_list_or_string_token ( bs, mal->value->ext_language );
  if ( bs == NULL ) return NULL;

  return bs;
}
				  
/* Go horizontally across the bodystructure... and recurse in
 *  ( ... ) ( ... ) ( ... ) 
 */
char * rfc2060_decode_structure_list ( char * bs, 
				       MimeAttachmentList * mal )
{
  int
    type;

  MimeAttachmentList 
    * currNode = mal;
  
  while ( isspace ( *bs ) )
    bs++;
  
  while ( *bs++ == '(' ) { 
    if ( *bs == '(' ) { 
      /* Nested type, recurse in */
      currNode->leaf = alloc_mal();
      if ( currNode->leaf == NULL ) return NULL;
      bs = rfc2060_decode_structure_list ( bs,
					   currNode->leaf );
      if ( bs == NULL ) return NULL;      
    }
    else { 
      /* This is an actual mime attachment to decode, get the type */
      bs = rfc2060_get_type ( bs, &type );
      if ( type == MIME_TYPE_INVALID ) return NULL;
      switch ( type ) { 
      case MIME_TYPE_MESSAGE:
	bs = rfc2060_decode_message ( bs, currNode );
	break;
      case MIME_TYPE_MULTIPART:
	bs = rfc2060_decode_multipart ( bs, currNode );
	break;
      default:
	bs = rfc2060_decode_discrete ( bs, currNode, type );
	break;
      }
      if ( bs == NULL ) return NULL;
    }     
    while ( isspace ( *bs ) )
      bs++;
  }
  
  return bs;
}
		      
MimeAttachmentList * rfc2060_decode_bodystructure ( char * buffer )
{
  MimeAttachmentList
    * mal;

  mal = alloc_mal();
  if ( mal == NULL ) 
    return NULL;

  buffer = rfc2060_decode_structure_list ( buffer,
					   mal );

  return mal;  
}

void mal_print ( MimeAttachmentList * mal )
{
  if ( mal->value ) { 
    printf ( "Type: %d\nsubtype: %s\nparameter: %s\nid: %s\ndesc: %s\nencoding: %s\nsizebytes: %d\nsizelines: %d\next_param: %s\next_dispos: %s\next_lang: %s\n",
	     mal->value->type,
	     mal->value->subtype,
	     mal->value->id,
	     mal->value->description,
	     mal->value->encoding,
	     mal->value->sizeOctets,
	     mal->value->sizeLines,
	     mal->value->ext_parameter,
	     mal->value->ext_disposition,
	     mal->value->ext_language );
  }
  
  if ( mal->leaf ) 
    mal_print ( mal->leaf );

  if ( mal->next )
    mal-print ( mal->next );

}
char inbuffer[30000] = { 0 };

int main ( int argc, char ** argv ) 
{
  char * in;
  int i;
  MimeAttachmentList 
    * mal;

  in = inbuffer;
  while ( ( *in = getchar() ) != EOF )
    in++;
  *in = 0;

  printf ( "I read: %d bytes\n", strlen ( inbuffer ) );
  
  mal = rfc2060_decode_bodystructure ( inbuffer );
  
  return 0;
  
}
