/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#include <stdlib.h>
#include <limits.h>
#include "skip-list.h"

skiplist *newlist(int ml, float p, int (*cf)(const void *, const void *))
{
  skiplist *s = (skiplist *) xmalloc(sizeof(skiplist));
  int i;

  if (ml > SKIP_ABSMAXLVL) 
      ml = SKIP_ABSMAXLVL;
  s->maxlevel = ml;
  s->prob = p;
  s->comp = cf;
  s->items = 0;

  s->header = (skipnode *) xmalloc(sizeof(skipnode) + ml * sizeof(skipnode *));
  s->header->data = NULL;
  for (i = 0; i < ml; i++)
    s->header->forward[i] = NULL;

  for (i = 0; i < ml; i++)
    s->finger[i] = NULL;

  s->curlevel = 0;

  return s;
}

void *ssearch(skiplist *S, void *k)
{
  skipnode *x = S->header;
  int i;
  int lvl;

  lvl = 1;
  if (S->finger[0] != NULL) {	/* finger is valid */
      if ((S->finger[0] == S->header) || 
	  (S->comp(S->finger[0]->data, k) < 0)) {
	  /* move forward, find largest lvl such that
	     finger[lvl]->forward[lvl]->key < k */
	  while ((lvl <= S->curlevel) &&
		 (S->finger[lvl] && S->finger[lvl]->forward[lvl] &&
		  S->comp(S->finger[lvl]->forward[lvl]->data, k) < 0))
	      lvl++;
	  lvl--;
	  x = S->finger[lvl];
      } else {
	  /* move backwards, find the smallest lvl such that
	     finger[lvl]->key < k */
	  while ((lvl <= S->curlevel) &&
		 (S->finger[lvl] && S->comp(S->finger[lvl]->data, k) >= 0))
	      lvl++;
	  if (lvl > S->curlevel) {
	      lvl = S->curlevel;
	      x = S->header;
	  } else x = S->finger[lvl];
      }
  }	      

  /* invariant: x < k */
  for (i = lvl; i >= 0; i--)
      while ((x->forward[i] != NULL) && (S->comp(x->forward[i]->data, k) < 0))
	  x = x->forward[i];
  x = x->forward[0];
  if ((x != NULL) && (S->comp(x->data, k) == 0))
      return x->data;
  else
      return NULL;
}

int randLevel(skiplist *S)
{
  int lvl = 0;
  while ((((float) rand() / (float) (INT_MAX)) < S->prob) 
	 && (lvl < S->maxlevel))
    lvl++;
  return lvl;
}

void sinsert(skiplist *S, void *k)
{
  skipnode *update[SKIP_ABSMAXLVL];
  int newlvl = randLevel(S), i;
  skipnode *x = S->header, 
           *new = (skipnode *) xmalloc(sizeof(skipnode) +
				      newlvl * sizeof(skipnode *));
  
  for (i = S->curlevel; i >= 0; i--) {
    while (x->forward[i] && S->comp(x->forward[i]->data, k) < 0)
      x = x->forward[i];
    update[i] = x;
  }
  x = x->forward[0];
  if ((x != NULL) && (S->comp(x->data, k) == 0)) {
    x->data = k;
    free(new);
  } else {
    S->items++;
    if (newlvl > S->curlevel)
      for (i = S->curlevel + 1; i <= newlvl; i++)
	update[i] = S->header;
    new->data = k;
    for (i = 0; i <= newlvl; i++) {
      new->forward[i] = update[i]->forward[i];
      update[i]->forward[i] = new;
    }
  }

  for (i = 0; i <= S->maxlevel; i++)
      S->finger[i] = update[i];
}

void sdelete(skiplist *S, void *k)
{
  skipnode *update[SKIP_ABSMAXLVL];
  skipnode *x = S->header;
  int i;
 
  for (i = S->curlevel; i >= 0; i--) {
    while (x->forward[i] && S->comp(x->forward[i]->data, k) < 0)
      x = x->forward[i];
    update[i] = x;
  }
  x = x->forward[0];
  if ((x != NULL) && (S->comp(x->data, k) == 0)) {
    for (i = 0; i <= S->curlevel; i++) {
      if (update[i]->forward[i] != x) break;
      update[i]->forward[i] = x->forward[i];
    }
    free(x);
    S->items--;
    while ((S->curlevel >= 1) && (S->header->forward[S->curlevel] == NULL))
      S->curlevel--;
  }

  /* if it isn't here, do nothing */

  for (i = 0; i <= S->maxlevel; i++)
      S->finger[i] = update[i];
}

void destroylist(skiplist *S)
{
  skipnode *x = S->header, *y;

  while (x != NULL) {
    y = x->forward[0];
    free(x);
    x = y;
  }
  free(S);
}

int sempty(skiplist *S)
{
  return (S->items == 0);
}

void sforeach(skiplist *S, void (*f)(const void *))
{
  skipnode *s;

  s = S->header->forward[0];
  while (s != NULL) {
    f(s->data);
    s = s->forward[0];
  }
}

void *sfirst(skiplist *S, skipnode **ptr)
{
  *ptr = S->header->forward[0];
  if (*ptr != NULL)
    return (*ptr)->data;
  else
    return NULL;
}

void *snext(skipnode **ptr)
{
  if (*ptr != NULL)
    *ptr = (*ptr)->forward[0];
  if (*ptr != NULL)
    return (*ptr)->data;
  else
    return NULL;
}
