/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>

#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <netdb.h>

#include <sasl.h>
#include <saslutil.h>

#include <stdlib.h>

#include "smtp.h"
#include "text.h"
#include <prot.h>
#include "alias.h"
#include "xmalloc.h"
#include "console.h"
#include "clinet.h"
#include "ezconfig.h"
#include "parseaddr.h"

struct smtp_s {
    struct protstream *prot_in;
    struct protstream *prot_out;

    sasl_conn_t *saslconn;

    char *server;

    char *saslmechs;
};

static short g_smtp_port;

alias_t *g_aliases;

int smtp_init(short port)
{
    g_smtp_port = port;

    g_aliases = alias_init();

    return 0;
}

/* initialize the network */
static int init_net(char *serverFQDN, int port)
{
  struct sockaddr_in addr;
  struct hostent *hp;
  int sock;
  char   errorMessage[100];

  if ((hp = gethostbyname(serverFQDN)) == NULL) {
    snprintf ( errorMessage, 99, "Send failed: gethostbyname: %s\n", 
	       strerror ( errno ) );
    CON_WriteLine ( errorMessage );
    return -1;
  }

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {    
    snprintf ( errorMessage, 99, "Send failed: socket: %s\n", 
	       strerror ( errno ) );
    CON_WriteLine ( errorMessage );
    return -1;
  }

  addr.sin_family = AF_INET;
  memcpy(&addr.sin_addr, hp->h_addr, hp->h_length);
  addr.sin_port = htons(port);

  if (connect(sock, (struct sockaddr *) &addr, sizeof (addr)) < 0) {
    snprintf ( errorMessage, 99, "Send failed: connect: %s\n", 
	       strerror ( errno ) );
    CON_WriteLine ( errorMessage );
    return -1;
  }

  return sock;
}

int smtp_start(smtp_t **conn)
{
    int sock;
    const char *server;
    *conn = malloc(sizeof(smtp_t));
    if (!*conn) return -1;
    memset(*conn, 0, sizeof(smtp_t));

    /* connect to server */
    server = config_getstring("SmtpServer","smtp");
    sock = init_net((char *)server, g_smtp_port);
    if (sock == -1) return -1;

    (*conn)->server = xstrdup(server);
    (*conn)->prot_in  = prot_new(sock, 0);
    (*conn)->prot_out = prot_new(sock, 1);     
   
    return 0;
}

static int getcode(char *buf, int *lastone)
{
    int ret;

    if (strlen(buf)<4) return -1;

    if ((!isdigit((int) buf[0])) || (!isdigit((int) buf[0])) || (!isdigit((int) buf[0]))) {
	return -1;
    }

    ret = ((buf[0] - '0') * 100) +
	  ((buf[1] - '0') * 10) +
	  ((buf[2] - '0') * 1);

    if (lastone) {
	if (buf[3] == '-')
	    *lastone = 1;
	else
	    *lastone = 0;
    }

    return ret;
}


static int listen_initial(smtp_t *conn)
{
    char buf[4096];
    int num;
    int lastone;

    do {

	prot_fgets(buf, sizeof(buf)-1, conn->prot_in);

	num = getcode(buf,&lastone);
	if (num!=220) return -1;

    } while (lastone == 1);

    return 0;
}

static int say_hello(smtp_t *conn)
{
    char buf[4096];
    int num;
    int lastone;
    char host[MAXHOSTNAMELEN+1];
    int r;
    
    /* get our hostname */
    r = gethostname(host, sizeof(host)-1);
    if (r) {
	strcpy(host,"foo.bar");
    }

    prot_printf(conn->prot_out,"EHLO %s\r\n",host);
    prot_flush(conn->prot_out);

    do {
	char *cap;
	
	prot_fgets(buf, sizeof(buf)-1, conn->prot_in);

	num = getcode(buf,&lastone);
	if (num!=250) return -1;

	/* look for capabilities */
	cap = buf+4;
	if (strncasecmp(cap,"AUTH ",5)==0) {
	    cap[ strlen(cap)-2] = '\0';
	    conn->saslmechs = xstrdup(cap+5);
	}

    } while (lastone == 1);
    
    return 0;
}

static int write64(char *cout, int coutlen, struct protstream *prot_out)
{
    int r;
    char *enc = xmalloc(coutlen*2+5);
    int enclen;
    r = sasl_encode64(cout,coutlen,enc,coutlen*2+5,&enclen);
    if (r) return -1;
    
    prot_write(prot_out,enc,enclen);
    free(enc);	

    return 0;
}

static sasl_security_properties_t *make_secprops(int min,int max)
{
  sasl_security_properties_t *ret=(sasl_security_properties_t *)
  malloc(sizeof(sasl_security_properties_t));

  ret->maxbufsize=1024;
  ret->min_ssf=min;
  ret->max_ssf=max;

  ret->security_flags=0;
  ret->property_names=NULL;
  ret->property_values=NULL;

  return ret;
}

static int say_authenticate(smtp_t *conn)
{
    char buf[4096];
    int num;
    int lastone;
    sasl_conn_t *saslconn;
    int r;
    char *cout;
    unsigned coutlen;
    const char *mechusing;
    sasl_security_properties_t *secprops=NULL;
    int addrsize;
    struct sockaddr_in saddr_l;
    struct sockaddr_in saddr_r;
    const char *userid = config_getstring("Userid",NULL);

    sasl_callback_t cbs[5];

    cbs[0].id = SASL_CB_PASS;
    cbs[0].proc = &sasl_cb_getsecret;
    cbs[0].context = conn->server;
    cbs[1].id = SASL_CB_AUTHNAME;
    cbs[1].proc = &sasl_cb_getuser;
    cbs[1].context = (char *)userid;
    cbs[2].id = SASL_CB_USER;
    cbs[2].proc = &sasl_cb_getuser;
    cbs[2].context = (char *)userid;
    cbs[3].id = SASL_CB_LIST_END;

    r = sasl_client_new("smtp",
			conn->server,
			cbs,
			0, /* xxx secflags */
			&saslconn);

    if (r) {
	CON_WriteLine("Sasl error in sasl_client_new()");
	return -1;
    }

    conn->saslconn = saslconn;

    /* set options */
    secprops=make_secprops(0,128);
    if (secprops==NULL) return -1;

    r = sasl_setprop(saslconn, SASL_SEC_PROPS, secprops);
    if (r!=SASL_OK) return -1;
    free(secprops);

    addrsize=sizeof(struct sockaddr_in);
    if (getpeername(conn->prot_in->fd,(struct sockaddr *)&saddr_r,&addrsize)!=0)
	return -1;

    r = sasl_setprop(saslconn, SASL_IP_REMOTE, &saddr_r);
    if (r != SASL_OK) return -1;

    addrsize=sizeof(struct sockaddr_in);
    if (getsockname(conn->prot_in->fd,(struct sockaddr *)&saddr_l,&addrsize)!=0)
	return -1;

    r = sasl_setprop(saslconn,   SASL_IP_LOCAL, &saddr_l);
    if (r != SASL_OK) return -1;

    /* start authenticating */
    r = sasl_client_start(saslconn,
			  conn->saslmechs,
			  NULL,
			  NULL, /*sasl_interact_t **prompt_need,*/
			  &cout,
			  &coutlen,
			  &mechusing);

    if ((r!=SASL_OK) && (r!=SASL_CONTINUE)) {
	CON_WriteLine("Sasl error in sasl_client_start():");
	CON_WriteLine((char *) sasl_errstring(r,NULL,NULL));
	return -1;
    }
	    
    prot_printf(conn->prot_out,"AUTH %s",mechusing);
    if (coutlen > 0) {
	prot_printf(conn->prot_out," ");
	r = write64(cout,coutlen,conn->prot_out);
	if (r) return r;
	free(cout);
    }
    prot_printf(conn->prot_out,"\r\n");	    
    prot_flush(conn->prot_out);    

    while (1) {
	char *cin = NULL;
	unsigned cinlen = 0;
	

	/* get input */
	prot_fgets(buf, sizeof(buf)-1, conn->prot_in);	    
	num = getcode(buf,&lastone);

	if (num == 334) {
	    cin = xmalloc(strlen(buf));
	    
	    sasl_decode64(buf+4,strlen(buf+4),
			  cin,&cinlen);

	    r = sasl_client_step(saslconn,
				 cin, cinlen,
				 NULL,
				 &cout,&coutlen);

	    if ((r!=SASL_OK) && (r!=SASL_CONTINUE)) {
		CON_WriteLine("Sasl error in sasl_client_step():");
		CON_WriteLine((char *) sasl_errstring(r,NULL,NULL));
		return -1;
	    }

	    r = write64(cout,coutlen,conn->prot_out);
	    if (r) return r;
	    prot_printf(conn->prot_out,"\r\n");
	    prot_flush(conn->prot_out);
	    free(cout);
	    			  
	} else if (num == 235) {
	    
	    prot_setsasl(conn->prot_in, saslconn);
	    prot_setsasl(conn->prot_out, saslconn);
	    return 0;
	    
	} else {
	    CON_WriteLine("SMTP authentication error\n");
	    CON_WriteLine(buf+4);	    
	    return -1;
	}
    }
}

static int say_mailfrom(smtp_t *conn, char *mailfrom)
{
    char buf[4096];
    int num;
    int lastone;

    prot_printf(conn->prot_out,"MAIL FROM:<%s>\r\n",mailfrom);
    prot_flush(conn->prot_out);

    do {

	prot_fgets(buf, sizeof(buf)-1, conn->prot_in);

	num = getcode(buf,&lastone);
	if (num!=250) return -1;

    } while (lastone == 1);
    
    return 0;
}

static int say_rcptto(smtp_t *conn, char **rcptto, const char **errmsg)
{
    char buf[4096];
    int num;
    int lastone;
    int cur = 0;

    while (rcptto[cur]) {

	alias_smart_add(g_aliases,rcptto[cur]);

	prot_printf(conn->prot_out,"RCPT TO:<%s>\r\n",rcptto[cur]);
	prot_flush(conn->prot_out);

	do {
	    
	    prot_fgets(buf, sizeof(buf)-1, conn->prot_in);
	    
	    num = getcode(buf,&lastone);
	    if (num!=250) {
		*errmsg = xstrdup(buf);
		return -1;
	    }
	    
	} while (lastone == 1);

	cur++;
    }
    
    return 0;
}

static int say_data(smtp_t *conn)
{
    char buf[4096];
    int num;
    
    prot_printf(conn->prot_out,"DATA\r\n");
    prot_flush(conn->prot_out);

    prot_fgets(buf, sizeof(buf)-1, conn->prot_in);

    num = getcode(buf,NULL);
    if (num!=354) return -1;
    
    return 0;
}

/*
 * Expand address with alias lookups and defaultdomains. Returned
 * string must be malloced
 */
static char *
expand_address(struct address *addr)
{
    char *alias;
    char * final_address;
    int len;
    char *domain = addr->domain;

    if ((!addr->domain) || (strcasecmp(addr->domain,"unspecified-domain") == 0)) {
	char * default_domain;

	alias = alias_lookup(g_aliases, addr->mailbox);
	if (alias) return xstrdup(alias);
	
	default_domain = (char *)config_getstring ( "DefaultDomain", "" );
	
	if ( default_domain == NULL || strlen ( default_domain ) == 0 ) {

	} else {	
	    domain = default_domain;
	}
    }

    /* JCS - +3 because we need the @, +, and the trailing NULL */
    len = strlen ( addr->mailbox);
    if (domain) len += strlen(domain);

    final_address = xmalloc (len+3);
	
    strcpy ( final_address, addr->mailbox );
    if (domain) {
	strcat ( final_address, "@" );
	strcat ( final_address, domain );
    }
	
    return final_address;
}

int parse_recipients_line(char *line, char **ret, int numret, char **end)
{
    char *s;
    char *start;
    char *endline;
    struct address *addrlist = NULL;
    struct address *addr;

    /* addresses seperated by commas */
    start = line;
    s = start;

    endline = strchr(s,'\n');
    if (endline) {
	*endline = '\0';
	if (end) *end = endline-1;
    } else {
	if (end) *end = line + strlen(line)-1;
    }

    parseaddr_list(s, &addrlist);

    addr = addrlist;

    while (addr) {

	ret[numret++] = expand_address(addr);
	ret[numret] = NULL;

	addr = addr->next;
    }

    parseaddr_free(addrlist);

    return numret;
}

static char **parse_recipients(TEXT_BUF *tbuf)
{
    char buf[4096];
    char *bufp;
    char **ret;
    int size;
    int bytes_read;

    bytes_read = TXT_GetData(tbuf,
			     buf, sizeof(buf)-1, 0);

    ret = malloc(sizeof(char *) * 20);
    
    if (strncmp(buf,"To:",3)!=0) {
	return NULL;
    }
    size = parse_recipients_line(buf+4,ret,0,&bufp);

    /* JCS - parse_recipients_line() leaves us at the charcater 
     * before the newline (turned NULL).  
     */
    bufp += 2;

    /* go past subject */
    bufp = strchr(bufp,'\n');
    if (!bufp) return ret;
    bufp++;

    if (strncmp(bufp,"Cc:",3)!=0) {
	return ret;
    }

    size = parse_recipients_line(bufp+4,ret,size,NULL);
    
    return ret;
}

static void
free_rcptto(char **rcptto)
{
    int i;

    if (rcptto==NULL) return;

    for (i=0;rcptto[i]!=NULL;i++) {
	free(rcptto[i]);
    }

    free(rcptto);
}

static int write_dot_stuffed(smtp_t *conn,
			     char *buf,
			     int len,
			     char *twol, char *lc)
{
    int i;
    char twolast = *twol;
    char lastchar = *lc;

    for(i=0;i<len;i++) {
	char ch = buf[i];

	/* do the dot stuff */
	if ((ch == '\n') && (lastchar == '.') && (twolast == '\n')) {
	    prot_putc('.', conn->prot_out);	    
	}

	/* LF -> CRLF */
	if (ch == '\n') prot_putc('\r', conn->prot_out);
	prot_putc(ch, conn->prot_out);
       
	
	twolast = lastchar;	
	lastchar = ch;

    }

    prot_flush(conn->prot_out);

    *twol = twolast;
    *lc = lastchar;
    
    return 0;
}

static int give_data(smtp_t *conn,
		     TEXT_BUF *tbuf)
{
    char buf[2048];
    int bytes_read;
    int tot_sent = 0;
    int num, i;
    char twolast = 'a';
    char lastchar = '\n';
    int linewidth_default = 74;
    int linelen;
    int linealreadywrote = 0;


    /* JCS - First send a From: line.  I know this is
     * a hack, but it makes email that I send look somewhat
     * more professional
     */

    strcpy ( buf, config_getstring ( "RealName", "" ) );
    prot_printf ( conn->prot_out, "From: " );
    if ( strlen ( buf ) != 0 ) {
      prot_putc ( '"', conn->prot_out );
      prot_printf ( conn->prot_out, buf );
      prot_printf ( conn->prot_out, "\" " );      
    }

    strcpy ( buf, config_getstring ( "Email", "" ) );
    prot_putc ( '<', conn->prot_out );
    prot_printf ( conn->prot_out, buf );
    prot_printf ( conn->prot_out, ">\r\n" );

    /* JCS - take the first three lines of the email and do not
     * add newlines, for they are headers and thy mailserver 
     * shall smite thou who put newlines in the middle of them.
     */

    i = 0;
    while ( i < 3 ) {
	bytes_read = TXT_GetData ( tbuf, buf, 1, tot_sent );

	if ( bytes_read > 0 && buf[i] != 0 ) {
	  tot_sent++;

	  if ( buf[0] == '\n' ) {
	    prot_putc ( '\r', conn->prot_out );
	    i++;
	  }
	  
	  prot_putc ( buf[0], conn->prot_out );	  
	  
	} else { 
	  break;
	}      
    }


    do {

	bytes_read = TXT_GetData(tbuf,
				 buf, sizeof(buf)-1, tot_sent);

	if (bytes_read > 0) {
	    int pos = 0;

	    do {
		linelen = TXT_FormatText(buf+pos, bytes_read-pos, linewidth_default - linealreadywrote);
		linealreadywrote = 0;
		
		write_dot_stuffed(conn,buf+pos,linelen,&twolast,&lastchar);

		pos += linelen;

		/*
		 * If last character wasn't a newline add one
		 */
		if ((lastchar != '\n') && (pos < bytes_read)) {
		    prot_putc('\r', conn->prot_out);
		    prot_putc('\n', conn->prot_out);
		}
	    
	    } while (pos < bytes_read);

	    if (lastchar != '\n') {
		linealreadywrote = linelen;
	    } else {
		linealreadywrote = 0;
	    }

	    tot_sent+=bytes_read;

	} else {
	    return -1;
	}

    } while (tot_sent < tbuf->fsize);

    prot_printf(conn->prot_out,"\r\n.\r\n");
    prot_flush(conn->prot_out);

    prot_fgets(buf, sizeof(buf)-1, conn->prot_in);

    num = getcode(buf,NULL);
    if (num!=250) return -1;

    return 0;
}

int smtp_send(smtp_t *conn, char *mailfrom,
	      TEXT_BUF *tbuf)
{
    int r;
    char **rcptto;
    const char *errmsg;

    if (!conn) return -1;
    if (!conn->prot_out) return -1;

    r = listen_initial(conn);
    if (r) {
	CON_WriteLine("Error connecting to SMTP server");
	return -1;    
    }

    /* say hello */
    r = say_hello(conn);
    if (r) {
	CON_WriteLine("Error saying hello to SMTP server");
	return -1;
    }

    /* if server listed sasl mechs assume we have to authenticate */
    if (conn->saslmechs) {
	r = say_authenticate(conn);
	if (r) {
	    CON_WriteLine("Couldn't SMTP authenticate but trying to send message anyway");
	}
    }

    /* mail from */
    r = say_mailfrom(conn, mailfrom);
    if (r) {
	CON_WriteLine("Error saying mail from to SMTP server");
	return -1;
    }

    rcptto = parse_recipients(tbuf);

    /* rcpt to */
    r = say_rcptto(conn,rcptto,&errmsg);
    if (r) {
	CON_WriteLine("Error doing rcpt to to SMTP server");
	CON_WriteLine((char *)errmsg);
	free_rcptto(rcptto);
	return -1;
    }
    free_rcptto(rcptto);

    /* data cmd */
    r = say_data(conn);
    if (r) {
	CON_WriteLine("Error doing data command to SMTP server");
	return -1;
    }

    /* actual data */
    r = give_data(conn, tbuf);
    if (r) {
	CON_WriteLine("Error giving message data to SMTP server");
	return -1;
    }

    return 0;
}

void smtp_close(smtp_t *conn)
{
    if (!conn) return;

    if (conn->prot_out) {
	prot_printf(conn->prot_out,"QUIT\r\n");
	prot_flush(conn->prot_out);

	prot_free(conn->prot_in);
	prot_free(conn->prot_out);
    }

    if (conn->server) {
	free(conn->server);
    }

    if (conn->saslmechs) {
	free(conn->saslmechs);
    }

    if (conn->saslconn) {
	sasl_dispose(&conn->saslconn);
	conn->saslconn = NULL;
    }

    free(conn);
}
