/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * windows.c -- window management module
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <curses.h>

#include "common.h"
#include "screen.h"
#include "comkeys.h"
#include "menus.h"
#include "and_keys.h"
#include "and_scr.h"
#include "windows.h"
#include "console.h"
#include "rli.h"
#include "list.h"
#include "rlci.h"

#ifdef IBMPC
#define DOUBLE_LINE_FRONT_WINDOW
#endif

#define MAX_IDS 255

#define MOVE_WINDOW 0
#define RESIZE_WINDOW 1

#define MIN_HEIGHT 5

extern unsigned char screen_rows, screen_cols;

Boolean g_WinRegionRefresh = FALSE;

int w_mode = W_NORMAL_MODE;

int w_ban_row = BAN_ROW;
int w_ban_col = BAN_COL;

int w_data_row = DATA_ROW;
int w_data_col = DATA_COL;

int w_height_offset = HEIGHT_OFFSET;
int w_width_offset = WIDTH_OFFSET;

static Boolean gHLSelection = TRUE;
static Boolean gCursorLock  = FALSE;

WINLIST *win_list = NULL;
WINLIST *win_unused = NULL;
WINLIST *win_list_end = NULL;

typedef unsigned char VIS_REGION;

VIS_REGION *vis_map = 0;
VIS_REGION *alt_vis_map = 0;

static Boolean write_liberally = TRUE;

static int refresh_region_x;
static int refresh_region_y;
static int refresh_region_w;
static int refresh_region_h;

void (*menu_setup_proc)() = 0;

void WIN_SetMode(int mode)
{
  switch (mode) {
  case W_NOBORD_MODE:
    w_ban_col = 0;
    w_data_col = 0;
    w_width_offset = 0;
    w_height_offset = HEIGHT_OFFSET - 2;
    w_mode |= W_NOBAN_MODE;
  case W_NOBAN_MODE:
    if (w_mode & W_NOBORD_MODE) break;
    if (mode & W_NOBAN_MODE) w_height_offset = HEIGHT_OFFSET - 1;
    w_data_row = DATA_ROW - 1;
  case W_ZOOM_ALL:
    if (w_mode & W_NOBORD_MODE) break;
    w_mode |= mode;
    WIN_AdjustSizes();
  }
}

void WIN_AdjustSizes(void)
{
  if (win_list && (w_mode & ~W_NORMAL_MODE)) {
    WINLIST *awp;
    for (awp = win_list; awp != NULL; awp = awp->next) {
      WINFO *w = awp->awin;
      if (w_mode & (W_NOBORD_MODE | W_ZOOM_ALL)) {
	if (w_mode & W_ZOOM_ALL) WIN_Zoom(awp);
	else {
	  w->loc.x = 0;
	  w->loc.y = 1;
	  w->dim.x = screen_cols;
	  w->dim.y = screen_rows - w->loc.y - 1;
	}
      }
      WIN_WindowChangedSize(w);
    }
    WIN_CalcVisRegion();
  }
}

void WIN_init(void (*mproc)())
{
  menu_setup_proc = mproc;
  if (vis_map == 0) {
    vis_map = (VIS_REGION *)malloc((screen_rows + 1) * screen_cols);
    alt_vis_map = (VIS_REGION *)malloc((screen_rows + 1) * screen_cols);
    if (vis_map == (VIS_REGION *)NULL || alt_vis_map == (VIS_REGION *)NULL) {
      CON_WriteLine(MSG_NotEnoughMemory);
      exit(-1);
    }
  }
  WIN_CalcVisRegion();
}

WINLIST *WIN_CreateWindow(row, col, height, width,
			  hproc, dokey, draw, close_proc,tp)
int row, col, height, width;
void (*hproc)();
int  (*dokey)();
void (*draw)();
void (*close_proc)();
int tp;
{
  WINFO *awin;

  if ((awin = (WINFO *)malloc(sizeof(WINFO)))) {
    awin->id = WIN_GetNewId();
    awin->loc.y = row;
    awin->loc.x = col;
    awin->dim.y = height;
    awin->dim.x = width;
    awin->data_height = imax(0, height - w_height_offset);
    awin->data_width  = imax(0, width  - w_width_offset);
    awin->zoom  = FALSE;
    awin->shown = FALSE;
    awin->hproc = hproc;
    awin->dokey = dokey;
    awin->draw  = draw;
    awin->close_proc = close_proc;
    awin->title[0] = '\0';

    awin->window_type = tp;

    if (WIN_link(awin) == 0) {
      WIN_CalcVisRegion();
      return(win_list_end);
    }
    free(awin);
  }
  return((WINLIST *)NULL);
}

WINLIST *WIN_Create(row, col, height, width, hproc, dokey, draw, close_proc,tp)
int row, col, height, width;
void (*hproc)();
int  (*dokey)();
void (*draw)();
void (*close_proc)();
int tp;
{
  if (w_mode & (W_NOBORD_MODE | W_ZOOM_ALL)) {
    row = 1;
    col = 0;
    height = screen_rows - row - 1;
    width  = screen_cols;
  }

  return(WIN_CreateWindow(row,
			  col,
			  height,
			  width,
			  hproc,
			  dokey,
			  draw,
			  close_proc,
			  tp)
	 );
}

int WIN_link(WINFO *awin)
{
  WINLIST *alink;

  if ((alink = (WINLIST *)malloc(sizeof(WINLIST)))) {
    alink->awin = awin;
    WIN_LinktoTail(alink);
  }
  return(alink ? 0 : -1);
}

void WIN_LinktoHead(WINLIST *alink)
{
  if (win_list) win_list->prev = alink;
  alink->prev = (WINLIST *)NULL;
  alink->next = win_list;
  win_list = alink;
  if (win_list_end == (WINLIST *)NULL) win_list_end = win_list;
}

void WIN_LinktoTail(WINLIST *alink)
{
  if (win_list_end) win_list_end->next = alink;
  alink->prev = win_list_end;
  alink->next = (WINLIST *)NULL;
  win_list_end = alink;
  if (win_list == (WINLIST *)NULL) win_list = win_list_end;
}

void WIN_dlink(WINLIST *alink,int free_link)
{
  if (alink->prev) alink->prev->next = alink->next;
  if (alink->next) alink->next->prev = alink->prev;

  if (win_list == alink) win_list = alink->next;
  if (win_list_end == alink) win_list_end = alink->prev;

  if (alink == win_list) win_list = alink->next;

  if (alink && free_link) {
    if (alink->awin) free(alink->awin);
    free(alink);
  }
}

void WIN_Dispose(WINLIST *alink)
{
  if (WIN_isShown(alink)) WIN_HideWindow(alink);
  WIN_dlink(alink, TRUE);
}

void WIN_BehindFrontWindow(WINLIST *awp)
{
  WINFO *w = awp->awin;
  WINLIST *front_window = win_list;

  if (front_window == awp) return;

  WIN_dlink(front_window, FALSE);
  WIN_dlink(awp, FALSE);

  WIN_LinktoHead(awp);
  WIN_LinktoHead(front_window);

  WIN_CalcVisRegion();
  WIN_RefreshRegion(w->loc.y, w->loc.x, w->dim.y, w->dim.x);
}

void WIN_BringToFront(WINLIST *alink)
{
  Boolean do_hproc = FALSE;
  WINLIST *cur_win = win_list;
  
  if (alink != win_list) {
    WIN_dlink(alink, FALSE);
    WIN_LinktoHead(alink);
    do_hproc = TRUE;
  }

  WIN_OverlayMap(win_list->awin);

#ifdef DOUBLE_LINE_FRONT_WINDOW
  if (gHLSelection && cur_win && cur_win != alink) {
    WIN_Box(cur_win, cur_win->awin->dim.y, cur_win->awin->dim.x, 0, 0, NORMAL);
    WIN_SetTitle(cur_win, NIL);
  }
#endif

  WIN_Box(win_list, win_list->awin->dim.y,win_list->awin->dim.x, 0, 0, NORMAL);

  WIN_SetTitle(win_list, NIL);

  if (do_hproc) {
    if (gHLSelection) {
      if (cur_win && WIN_isShown(cur_win))
	if (cur_win->awin->hproc) (*cur_win->awin->hproc)(cur_win);
    }
    if (win_list->awin->hproc) (*win_list->awin->hproc)(win_list);
  }

  if (win_list->awin->draw) (*win_list->awin->draw)(win_list);
  WIN_DoSetupMenus();
}

void WIN_DoSetupMenus(void)
{
  if (menu_setup_proc) (*menu_setup_proc)();
  if (w_mode & W_NOBORD_MODE) {
    MNU_EnableItem(M_MOVE, FALSE);
    MNU_EnableItem(M_RESIZE, FALSE);
    MNU_EnableItem(M_ZOOM, FALSE);
  }
}

int WIN_FrontWindow(WINLIST *alink)
{
  if (!gHLSelection) return(FALSE);
  return((win_list == alink && WIN_isShown(win_list)) ? TRUE : FALSE);
}

void WIN_RefreshRegion(int r,int c,int h,int w)
{
  int i,j, idx, cursor_off;
  WINLIST *alink;
  Boolean win_ids[MAX_IDS];
  VIS_REGION *save_vis_map = vis_map;

  cursor_off = SCR_CursorOff();

  vis_map = alt_vis_map;
  bzero(vis_map, screen_rows * screen_cols);
  bzero(win_ids, sizeof(win_ids));

  write_liberally = FALSE;

  g_WinRegionRefresh = TRUE;

  for (i=r; i<r+h && i <= screen_rows; i++) {
    for (j=c; j<c+w && j < screen_cols; j++) {
      idx = i * screen_cols + j;
      if (save_vis_map[idx] != 0) {
	vis_map[idx] = save_vis_map[idx];
	win_ids[vis_map[idx]] = TRUE;
      } else {
	SCR_Write(NIL, i, j, 1, NORMAL);
      }
    }
  }

  for (alink = win_list_end; alink!=NULL; alink = alink->prev) {
    if (WIN_isShown(alink) && win_ids[alink->awin->id])
      WIN_Box(alink, alink->awin->dim.y, alink->awin->dim.x, 0, 0, NORMAL);
  }

  for (alink = win_list_end; alink!=NULL; alink = alink->prev) {
    if (WIN_isShown(alink) && win_ids[alink->awin->id]) {
      if (alink->awin->hproc) (*alink->awin->hproc)(alink);
      if (alink->awin->title[0]) WIN_SetTitle(alink, NIL);
    }
  }

  for (alink = win_list_end; alink!=NULL; alink = alink->prev) {
    if (WIN_isShown(alink) && win_ids[alink->awin->id])
      if (alink->awin->draw) (*alink->awin->draw)(alink);
  }

  SCR_UpdateScreen();

  vis_map = save_vis_map;
  write_liberally = TRUE;

  g_WinRegionRefresh = FALSE;

  if (cursor_off) SCR_CursorOn();
}

void WIN_Write(WINLIST *alink,char *text,int row, int col, int len, int attr)
{
  if (!WIN_isShown(alink)) return;
  row += alink->awin->loc.y;
  col += alink->awin->loc.x;

  if (write_liberally && (alink == win_list))
    SCR_Write(text, row, col, len, attr);
  else {
    WINFO *w = alink->awin;
    int id = w->id;
    while (len--) {
      if (row >= screen_rows || col >= screen_cols) break;
      if (vis_map[row*screen_cols+col] == id)
	SCR_Write(text, row, col, 1, attr);
      col++;
      if (text && *text) text++;
    }
  }
}

void WIN_Update(WINLIST *alink)
{
  if (write_liberally && WIN_isShown(alink))
    SCR_UpdateScreen();
}

void WIN_HideWindow(WINLIST *aw)
{
  if (WIN_isShown(aw)) {
    WINFO *w = aw->awin;

    if (aw != win_list_end) {
      WIN_dlink(aw, FALSE);
      WIN_LinktoTail(aw);
    }
    WIN_Hide(aw);
    WIN_CalcVisRegion();

    WIN_RefreshRegion(w->loc.y, w->loc.x, w->dim.y, w->dim.x);

#ifdef DOUBLE_LINE_FRONT_WINDOW
    if (gHLSelection) WIN_BringToFront(win_list);
#endif

#ifdef DO_HIGHLIGHT_SELECTION
    if (win_list->awin->hproc) (*win_list->awin->hproc)(win_list);
#endif
    WIN_DoSetupMenus();
  }
}

void WIN_OverlayMap(WINFO *win)
{
  register int i, j;
  if (!win->shown) return;
  for (i=win->loc.y; i<win->loc.y+win->dim.y && i <= screen_rows; i++)
    for (j=win->loc.x; j<win->loc.x+win->dim.x && j < screen_cols; j++)
      vis_map[i*screen_cols+j] = win->id;
}

void WIN_CalcVisRegion(void)
{
  WINLIST *awp;
  bzero(vis_map, screen_rows * screen_cols);

  if ((w_mode & W_NOBORD_MODE) || (win_list && win_list->awin->zoom))
    WIN_OverlayMap(win_list->awin);
  else {
    for (awp = win_list_end; awp != NULL; awp = awp->prev)
      WIN_OverlayMap(awp->awin);
  }
}

void WIN_RedrawScreen(void)
{
  MNU_SetupMenuBar();
  WIN_RefreshRegion(1, 0, screen_rows-1, screen_cols);
  WIN_DoSetupMenus();
  CON_RedrawBusyState();
}

int WIN_DoKey(int key)
{
  if (gNovice == FALSE) SCR_ClearStatusLine();

  switch (key) {
  case KEYCODE_REDRAW:
    SCR_RedrawScreen();
    break;
  default:
    if (win_list->awin->dokey)
      return((*win_list->awin->dokey)(key));
    return(FALSE);
  }

  return(TRUE);
}

int WIN_GetNewId(void)
{
#define START_ID 1

  if (win_list == NULL)
    return(START_ID);
  else {
    int i;
    WINLIST *awp;
    Boolean map[MAX_IDS];

    /* catch 1 length loop */
    if (win_list == win_list->next) {
	win_list->next = NULL;
    }

    bzero(&map[0], MAX_IDS);
    for (awp=win_list; awp != NULL; awp=awp->next) {
      map[awp->awin->id] = TRUE;
    }
    for (i=START_ID; i<MAX_IDS; i++)
      if (map[i] == FALSE) return(i);
  }

  return 0;
}

void WIN_LockCursor(int lock)
{
  gCursorLock = lock;
  if (!lock) WIN_HomeCursor();
}

void WIN_HomeCursor(void)
{
  if (!gCursorLock) {
#ifndef DO_HIGHLIGHT_SELECTION
    if (win_list)
      SCR_SetCursorPosn(win_list->awin->loc.y, win_list->awin->loc.x);
    else 
      SCR_SetCursorPosn(0, 0);
#else
    SCR_SetCursorPosn(0, 0);
#endif
  }
  if (write_liberally) SCR_UpdateScreen();
}

void WIN_Move(WINLIST *awp)
{
  WINFO *win = awp->awin;
  int r = win->loc.y;
  int c = win->loc.x;
  int h = win->dim.y;
  int w = win->dim.x;

  if (WIN_MoveOrResize(win, MOVE_WINDOW)) {
    WIN_CalcVisRegion();
    WIN_RefreshUnion(r, c, h, w, win);
  }
}

void WIN_Zoom(WINLIST *awp)
{
  WINFO *w = awp->awin;

  if (w->zoom) {
    w->loc.y = w->zoomloc.y;
    w->loc.x = w->zoomloc.x;
    w->dim.y = w->zoomdim.y;
    w->dim.x = w->zoomdim.x;
  } else {
    w->zoomloc.y = w->loc.y;
    w->zoomloc.x = w->loc.x;
    w->zoomdim.y = w->dim.y;
    w->zoomdim.x = w->dim.x;
    w->loc.y = 1;
    w->loc.x = 0;
    w->dim.y = screen_rows - w->loc.y - 1;
    w->dim.x = screen_cols - w->loc.x;
  }

  WIN_WindowChangedSize(w);

  w->zoom = (w->zoom ? FALSE : TRUE);
  if (WIN_isShown(awp)) {
    WIN_CalcVisRegion();
    WIN_RefreshRegion(1, 0, screen_rows - 1, screen_cols);
  }
}

void WIN_Resize(WINLIST *awp)
{
  WINFO *win = awp->awin;
  int r = win->loc.y;
  int c = win->loc.x;
  int h = win->dim.y;
  int w = win->dim.x;

  if (WIN_MoveOrResize(win, RESIZE_WINDOW)) {
    WIN_WindowChangedSize(win);
    WIN_CalcVisRegion();
    WIN_RefreshUnion(r, c, h, w, win);
  }
}

void WIN_WindowChangedSize(WINFO *win)
{
  win->data_height = imax(0, win->dim.y - w_height_offset);
  win->data_width  = imax(0, win->dim.x - w_width_offset);
}

void WIN_RefreshUnion(int r,int c,int h,int w,WINFO *win)
{
  h = imax(r+h, win->loc.y+win->dim.y);
  w = imax(c+w, win->loc.x+win->dim.x);
  r = imin(r, win->loc.y);
  c = imin(c, win->loc.x);
  h -= r;
  w -= c;

  WIN_RefreshRegion(r, c, h, w);
}

int WIN_MoveOrResize(WINFO *win,int move_or_resize)
{
  int key;
  int r, c, h, w, *rp, *cp;
  r = win->loc.y;
  c = win->loc.x;
  h = win->dim.y;
  w = win->dim.x;

  if (gNovice) SCR_WriteStatusLine(MSG_MoveOrResizeHint);

  if (move_or_resize == MOVE_WINDOW) {
    rp = &r;
    cp = &c;
    SCR_SetCursorPosn(r, c);
  } else { /* RESIZE_WINDOW */
    rp = &h;
    cp = &w;
    SCR_SetCursorPosn(r+h-1, c+w-1);
  }
  SCR_UpdateScreen();
  while ((key = SCR_GetKey())) {
    switch (key) {
    case KEYCODE_UP:
      *rp -= 1;
      if (WIN_CheckBounds(r, c, h, w)) break;
      *rp += 1; continue;
    case KEYCODE_DOWN:
      *rp += 1;
      if (WIN_CheckBounds(r, c, h, w)) break;
      *rp -= 1; continue;
    case KEYCODE_RIGHT:
      *cp += 1;
      if (WIN_CheckBounds(r, c, h, w)) break;
      *cp -= 1; continue;
    case KEYCODE_LEFT:
      *cp -= 1;
      if (WIN_CheckBounds(r, c, h, w)) break;
      *cp += 1; continue;
    case KEYCODE_RETURN:
    case KEYCODE_FAKERETURN:
      win->loc.y = r;
      win->loc.x = c;
      win->dim.y = h;
      win->dim.x = w;
      return(TRUE);
    case KEYCODE_ESCAPE:
      return(FALSE);
    default:
      SCR_Beep();
      continue;
    }
    if (move_or_resize == MOVE_WINDOW)
      SCR_SetCursorPosn(r, c);
    else /* RESIZE_WINDOW */
      SCR_SetCursorPosn(r+h-1, c+w-1);
    SCR_UpdateScreen();
  }
  return(FALSE);
}

int WIN_CheckBounds(int r, int c, int h, int w)
{
  if ((h < MIN_HEIGHT) ||
      (w < MIN_WIDTH) ||
      (r < 1) ||
      (c < 0) ||
      (c+w > screen_cols) ||
      (r+h >= screen_rows)) return(FALSE);

  return(TRUE);
}

void WIN_CloseFrontWindow(void)
{
  if (win_list) {
    /*
      if window has a close proc, let them handle it,
      else just hide the window.  Hope there is a SHOW for it!
    */
    if (win_list->awin->close_proc)
      (*win_list->awin->close_proc)(win_list);
    else
      WIN_HideWindow(win_list);
  }
}

static WINLIST *_gl_winlist = (WINLIST *)NULL;

void WIN_BoxDrawProc(unsigned char *data,int row, int col, int len, int attr)
{
  WIN_Write(_gl_winlist, data, row, col, len, attr);
}

void WIN_Box(WINLIST *awin,int rows, int cols, int ypos, int xpos, int attr)
{
  _gl_winlist = awin;

  if (w_mode & W_NOBORD_MODE)
    SCR_Line(1, cols, 0, 0, NORMAL, WIN_BoxDrawProc);
  else {
    SCR_Box (rows, cols, ypos, xpos, attr, WIN_BoxDrawProc);

    if ((w_mode & W_NOBAN_MODE) == 0)
      SCR_Line(1, cols-2, ypos+BAN_ROW+1, xpos+1,
	       attr, WIN_BoxDrawProc);
  }

  _gl_winlist = (WINLIST *)NULL;
}

void WIN_SetTitle(WINLIST *awp,char *title)
{
  WINFO *w = awp->awin;
  int len;
  Boolean may_clear_old_title;
  int line_attr;

#ifdef DOUBLE_LINE_FRONT_WINDOW
  line_attr = (WIN_FrontWindow(awp) ? DOUBLE_LINE_BOX : SINGLE_LINE_BOX);
#else
  line_attr = SINGLE_LINE_BOX;
#endif

  may_clear_old_title = (w->title[0] ? TRUE : FALSE);

  if (title) {
    strncpy(w->title, title, MAX_BAN);
    w->title[MAX_BAN] = '\0';
  }

  _gl_winlist = awp;

  if (w->title[0] == '\0') {
    if (may_clear_old_title)
      SCR_Line(1, w->dim.x-2, 0, 1, NORMAL, WIN_BoxDrawProc);
  } else {
    len = imin(strlen(w->title), w->dim.x-MIN_WIDTH);
    SCR_Line(1, w->dim.x-2, 0, 1, NORMAL, WIN_BoxDrawProc);
    SCR_LeftT(0, (w->dim.x-len)/2, NORMAL, WIN_BoxDrawProc);
    SCR_RightT(0, (w->dim.x-len)/2+len+1, NORMAL, WIN_BoxDrawProc);
    WIN_Write(awp, w->title, 0, (w->dim.x-len)/2+1, len, HILITC);
  }

  _gl_winlist = (WINLIST *)NULL;
}

void WIN_HLSelection(int on_off)
{
#ifdef DO_HIGHLIGHT_SELECTION

  gHLSelection = on_off;
  if (win_list->awin->hproc)
    (*win_list->awin->hproc)(win_list);

#endif /* DO_HIGHLIGHT_SELECTION */
}

void WIN_DisplayPosition(WINLIST *awp,long offset, long size,int use_real_size)
{
   int num;
   char  bboard_name[STR255LEN];

  if (awp && WIN_isShown(awp)) {
    int pos, pos_len;
    char pos_buf[80], *pos_buf_pt = &pos_buf[0];
    if (size <= 0) pos = 0;
    else {
      int fpos = (offset * 100) / size;
      if (fpos == 0 && offset > 0) pos = 1; /* big file, small window */
      else pos = imax(0, imin(fpos, 100));
    }

    if (use_real_size) {
	/* JCS - get the size of the list, new messages and the title of the
	 *  bboard.  Note that the count of new messages is grossly inefficient
	 *  and ought to be done once (and correctly, at that... the way I do 
	 *  it is quite hackish right now -- counting the number of N's next
	 *  to a message).
	 */
	num = GET_LIST_SIZE(CAPTION_RL);
       
	RL_read ( FOLDER_RL, GET_LIST_SELECTION ( FOLDER_RL ), bboard_name );
	
	sprintf(pos_buf_pt, "%s (%d messages, %d new) [#%05ld]", 
		bboard_name, num, RL_count_new(), (long)(offset + 1));
    } else {
      if (pos <= 0) strcpy(pos_buf_pt, "[Top]");
      else if (pos >= 100) strcpy(pos_buf_pt, "[Bot]");
      else sprintf(pos_buf_pt, "[%02d%%]", pos);
    }
    pos_len = imin(awp->awin->data_width, strlen(pos_buf_pt));
    WIN_Write(awp, pos_buf_pt, w_ban_row,
	      imax(w_data_col, (w_data_col+awp->awin->data_width - pos_len)),
	      pos_len,
#ifdef DO_HIGHLIGHT_SELECTION
	      (WIN_FrontWindow(awp) ? WHTBLU : NORMAL)
#else
	      WHTBLU
#endif
	      );
  }
}

void WIN_Reinitialize(void)
{
  int nbytes = (screen_rows + 1) * screen_cols;
  VIS_REGION *nvis;
  refresh_region_x = 0;
  refresh_region_y = 0;
  refresh_region_w = screen_cols;
  refresh_region_h = screen_rows;

  if ((nvis = (VIS_REGION *)realloc(vis_map, nbytes))==(VIS_REGION *)NULL) {
      /*CON_LocalMsg(MSG_NotEnoughMemory, NIL);*/
  } else {
    vis_map = nvis;
  }
  WIN_CalcVisRegion();
  /*  TXT_ReallocScrLines();*/
}

